import contextlib
import os


@contextlib.contextmanager
def temp_env_vars(**kwargs):
    """
    Temporarily set the process environment variables.
    >>> with temp_env_vars(PLUGINS_DIR=u'test/plugins'):
    ...   "PLUGINS_DIR" in os.environ
    True
    >>> "PLUGINS_DIR" in os.environ
    """
    old_environ = dict(os.environ)
    kwargs = {k: str(v) for k, v in kwargs.items()}
    os.environ.update(**kwargs)
    try:
        yield
    finally:
        os.environ.clear()
        os.environ.update(old_environ)



import asyncio
import dataclasses
import itertools
from dataclasses import dataclass, fields, is_dataclass
from datetime import date, datetime
from enum import Enum
from time import time
from typing import (
    Any,
    AsyncGenerator,
    Dict,
    Generic,
    List,
    Literal,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    cast,
    get_args,
    get_origin,
)

from boto3.dynamodb.conditions import Attr, Key
from botocore.exceptions import ClientError
from mypy_boto3_dynamodb import DynamoDBClient
from mypy_boto3_dynamodb.service_resource import DynamoDBServiceResource, Table
from mypy_boto3_dynamodb.type_defs import (
    BatchGetItemOutputServiceResourceTypeDef,
    QueryOutputTableTypeDef,
    TransactGetItemTypeDef,
    UpdateItemOutputTableTypeDef,
    WriteRequestOutputTypeDef,
)

from nhs_aws_helpers import dynamodb, dynamodb_retry_backoff
from nhs_aws_helpers.common import (
    is_dataclass_instance,
    optional_origin_type,
    run_in_executor,
)
from nhs_aws_helpers.dynamodb_model_store.base_model import (
    BaseModel,
    model_properties_cache,
)

TBaseModel = TypeVar("TBaseModel", bound=BaseModel)
TBaseModel_co = TypeVar("TBaseModel_co", bound=BaseModel, covariant=True)
TModelKey = TypeVar("TModelKey", bound=Mapping[str, Any])
TPageItem = TypeVar("TPageItem")


@dataclass
class PagedItems(Generic[TPageItem]):
    items: List[TPageItem]
    last_evaluated_key: Optional[Dict[str, Any]]

    def __post_init__(self):
        self.items = self.items or []

    def __iter__(self):
        """this is to allow the dataclass to unpack as a tuple, not iterate the items"""
        yield self.items
        yield self.last_evaluated_key

    def __len__(self):
        return len(self.items)

    def __add__(self, other: "PagedItems") -> "PagedItems":
        items = []
        items.extend(self.items)
        items.extend(other.items)
        return PagedItems(items=items, last_evaluated_key=other.last_evaluated_key)

    def __iadd__(self, other: "PagedItems") -> "PagedItems":
        self.items.extend(other.items)
        self.last_evaluated_key = other.last_evaluated_key
        return self


class BaseModelStore(Generic[TBaseModel, TModelKey]):
    _base_model_type: Type[TBaseModel]
    _model_types: Dict[str, Type[TBaseModel]] = {}  # noqa: RUF012

    def __new__(cls, *args, **kwargs):
        if (
            not hasattr(cls, "_base_model_type")
            or not cls._base_model_type
            or not hasattr(cls._base_model_type, "model_key_fields")
        ):
            raise ValueError(f"{cls.__name__} needs a _base_model_type definition")

        instance = super().__new__(cls)
        return instance

    def __init__(
        self, table_name: str, model_type_index_name: Optional[str] = None, model_type_index_pk: Optional[str] = None
    ):
        self._table_key_fields = self._base_model_type.model_key_fields()
        self._partition_key = self._table_key_fields[0]
        self._table_name = table_name
        self._table: Optional[Table] = None
        self._service: Optional[DynamoDBServiceResource] = None
        self._client: Optional[DynamoDBClient] = None
        self._model_type_index_name = model_type_index_name
        self._model_type_index_pk = model_type_index_pk

    @classmethod
    def register_model_type(cls, model_type: Type[TBaseModel]):
        cls._model_types[model_type.__name__] = model_type

    @property
    def service(self) -> DynamoDBServiceResource:
        if self._service is None:
            self._service = dynamodb()
        return self._service

    @property
    def table(self) -> Table:
        if self._table is None:
            self._table = self.service.Table(self._table_name)
        return self._table

    @property
    def client(self) -> DynamoDBClient:
        return self.service.meta.client

    @property
    def table_key_fields(self) -> List[str]:
        return self._table_key_fields

    @classmethod
    def _deserialise_field(cls, field: dataclasses.Field, value: Any, **kwargs) -> Any:
        return cls.deserialise_value(field.type, value, **kwargs)

    @classmethod
    def deserialise_value(cls, value_type: type, value: Any, **kwargs) -> Any:  # noqa: C901
        value_type = optional_origin_type(value_type)

        if value and value_type in (bytes, bytearray) and hasattr(value, "value"):
            value = value.value
            if value_type is bytearray:
                value = bytearray(value)

        if value_type in (str, bool, bytes, bytearray):
            return value

        if is_dataclass(value_type):
            return cls.deserialise_model(value, value_type, **kwargs)

        if value_type in (int, float):
            return value_type(value)

        if value_type == datetime:
            return datetime.fromisoformat(value)

        if value_type == date:
            return date.fromisoformat(value)

        origin_type = get_origin(value_type)

        if origin_type is list:
            item_type = get_args(value_type)[0]
            return [cls.deserialise_value(item_type, val, **kwargs) for val in value]

        if origin_type is dict:
            val_type = get_args(value_type)[1]
            return {key: cls.deserialise_value(val_type, val, **kwargs) for key, val in value.items()}

        if origin_type is frozenset:
            return frozenset(val for val in value)

        if origin_type is set:
            return set(value)

        if issubclass(value_type, Enum):
            return value_type(value)

        return value

    @classmethod
    def _before_deserialise_model(
        cls, model_dict: Dict[str, Any], model_type: Type[TBaseModel], **kwargs
    ) -> Dict[str, Any]:
        return model_dict

    @classmethod
    def deserialise_model(
        cls, model_dict: Dict[str, Any], model_type: Type[TBaseModel_co], **kwargs
    ) -> Optional[TBaseModel_co]:
        if model_dict is None:
            return None

        if not is_dataclass(model_type):
            raise TypeError(f"type {model_type} is not a dataclass")

        model_dict = cls._before_deserialise_model(model_dict, cast(Type[TBaseModel], model_type), **kwargs)

        model_fields = fields(model_type)

        deserialised: Dict[str, Any] = {}
        for field in model_fields:
            value = model_dict.get(field.name)
            if value is None:
                continue

            deserialised[field.name] = cls._deserialise_field(field, value, **kwargs)

        return model_type(**deserialised)  # type: ignore[return-value]

    @classmethod
    def serialise_value(cls, value: Any, **kwargs) -> Optional[Any]:
        if value is None:
            return None

        if is_dataclass_instance(value):
            class_serialised = cls.serialise_model(value, **kwargs)
            return class_serialised

        if isinstance(value, dict):
            dict_serialised = {k: cls.serialise_value(v, **kwargs) for k, v in value.items()}
            return dict_serialised

        if isinstance(value, (list, tuple)):
            if not value:
                return None
            list_serialised = [cls.serialise_value(v, **kwargs) for v in value]
            return list_serialised

        if isinstance(value, (date, datetime)):
            value = value.isoformat()

        if isinstance(value, (set, frozenset)) and not value:
            return None

        return value

    @classmethod
    def _serialise_field(
        cls, model: Any, field_name: str, field_type: type, metadata: Mapping[str, Any], value: Any, **kwargs
    ) -> Any:
        if value is None:
            return None

        if field_type == Optional[str] and value == "":
            return None

        value = cls.serialise_value(value, **kwargs)
        return value

    @classmethod
    def serialise_model(cls, model: Any, **kwargs) -> Optional[Dict[str, Any]]:
        if model is None:
            return None

        if not is_dataclass_instance(model):
            raise TypeError(f"type {type(model)} is not a dataclass")

        model_fields = model_properties_cache(model.__class__)

        result: Dict[str, Any] = {}

        for field_name, field_type, metadata in model_fields:
            value = getattr(model, field_name)

            value = cls._serialise_field(model, field_name, field_type, metadata, value, **kwargs)

            if value is None:
                # don't store None values.
                continue

            result[field_name] = value

        return result

    @dynamodb_retry_backoff()
    async def get_item_with_retry_info(
        self, key: TModelKey, **kwargs
    ) -> Tuple[Optional[Dict[str, Any]], float, str, int]:
        kwargs["Key"] = cast(Dict[str, str], key)

        started = time()
        response = await run_in_executor(self.table.get_item, **kwargs)
        duration = time() - started

        aws_request_id = response["ResponseMetadata"]["RequestId"]
        aws_retries = response["ResponseMetadata"]["RetryAttempts"]

        item: Optional[Dict[str, Any]] = response.get("Item")
        if not item:
            return None, duration, aws_request_id, aws_retries

        return item, duration, aws_request_id, aws_retries

    async def get_item(self, key: TModelKey, **kwargs) -> Optional[Dict[str, Any]]:
        item, _, _, _ = await self.get_item_with_retry_info(key, **kwargs)
        return item

    async def get_model_with_retry_info(
        self, key: TModelKey, model_type: Type[TBaseModel_co], **kwargs
    ) -> Tuple[Optional[TBaseModel_co], float, str, int]:
        item, duration, aws_request_id, aws_retries = await self.get_item_with_retry_info(key, **kwargs)

        if not item:
            return None, duration, aws_request_id, aws_retries

        return (
            cast(Optional[TBaseModel_co], self.deserialise_model(item, model_type, **kwargs)),
            duration,
            aws_request_id,
            aws_retries,
        )

    async def get_model(self, key: TModelKey, model_type: Type[TBaseModel_co], **kwargs) -> Optional[TBaseModel_co]:
        model, _, _, _ = await self.get_model_with_retry_info(key, model_type, **kwargs)

        return model

    async def transact_get_model(
        self, key: TModelKey, model_type: Type[TBaseModel_co], **kwargs
    ) -> Optional[TBaseModel_co]:
        items = await self.transact_get_items(cast(List[TransactGetItemTypeDef], [{"Get": {"Key": key}}]))

        if not items or not items[0]:
            return None

        item = cast(dict, items[0])

        return cast(Optional[TBaseModel_co], self.deserialise_model(item, model_type, **kwargs))

    @dynamodb_retry_backoff()
    async def transact_get_items(self, get_items: Sequence[TransactGetItemTypeDef]) -> List[Optional[dict]]:
        get_items = list(get_items)
        for get_item in get_items:
            get_item["Get"]["TableName"] = self._table_name
        response = await run_in_executor(self.client.transact_get_items, TransactItems=get_items)

        responses = response.get("Responses")
        if not responses:
            return []

        items = [resp.get("Item", None) for resp in responses]
        return items

    def _model_key_tuple(self, key_or_item: Mapping[str, Any]) -> Tuple[str, ...]:
        return tuple(key_or_item[field] for field in self._base_model_type.model_key_fields())

    async def transact_get_models(self, keys: Sequence[TModelKey], **kwargs) -> List[Optional[TBaseModel]]:
        keys = list(keys)
        key_indexes = {self._model_key_tuple(key): ix for ix, key in enumerate(keys)}

        records = await self.transact_get_items([cast(TransactGetItemTypeDef, {"Get": {"Key": key}}) for key in keys])

        models: List[Tuple[Tuple[str, ...], TBaseModel]] = [
            (
                self._model_key_tuple(record),
                cast(TBaseModel, self.deserialise_model(record, self._model_types[record["model_type"]], **kwargs)),
            )
            for record in records
            if record
        ]
        results: List[Optional[TBaseModel]] = [None for _ in range(len(keys))]

        unexpected = []

        for key_tuple, model in models:
            ix = key_indexes.get(key_tuple)
            if ix is None:
                unexpected.append(model)
                continue
            results[ix] = model

        if unexpected:
            raise ValueError(f"unexpected items in batch result: {[model.model_key() for model in unexpected]}")

        return results

    @dynamodb_retry_backoff()
    async def put_item(self, item: Dict[str, Any], **kwargs):
        kwargs["Item"] = item
        await run_in_executor(self.table.put_item, **kwargs)

    async def put_model(self, model: TBaseModel, **kwargs):
        assert model
        item = self.serialise_model(model, **kwargs)
        assert item
        await self.put_item(item, **kwargs)

    @dynamodb_retry_backoff()
    async def put_item_if_not_exists(self, item: Dict[str, Any], **kwargs) -> bool:
        try:
            await self.put_item(item, ConditionExpression=Attr(self._partition_key).not_exists(), **kwargs)
            return True
        except ClientError as e:
            # Ignore the ConditionalCheckFailedException and return false
            if e.response.get("Error", {}).get("Code", "") == "ConditionalCheckFailedException":
                return False
            # other exceptions bubble up
            raise

    async def put_model_if_not_exists(self, model: TBaseModel, **kwargs) -> bool:
        assert model
        item = self.serialise_model(model, **kwargs)
        assert item
        return await self.put_item_if_not_exists(item, **kwargs)

    async def item_exists(self, key: TModelKey, consistent_read: bool = False, **kwargs) -> bool:
        """
        Note: don't call this method as a check before doing a put item, as it will result
        in 2 hits to the DB instead of 1 and won't ensure consistency anyway.

        Instead, just call _put_item with an attribute_not_exists condition.
        """

        item = await self.get_item(
            key, ConsistentRead=consistent_read, ProjectionExpression=self._partition_key, **kwargs
        )
        return not (not item or self._partition_key not in item)

    @dynamodb_retry_backoff()
    async def delete_item(self, key: TModelKey, **kwargs):
        await run_in_executor(self.table.delete_item, Key=cast(Dict[str, Any], key), **kwargs)

    @dynamodb_retry_backoff()
    async def paginate(
        self, paginator_type: Literal["query", "scan", "list_backups", "list_tables"], **kwargs
    ) -> AsyncGenerator[dict, None]:
        paginator = self.client.get_paginator(paginator_type)
        kwargs["TableName"] = self._table_name
        page_iterator = paginator.paginate(**kwargs).__iter__()

        @dynamodb_retry_backoff()
        def _get_next_page():
            try:
                return page_iterator.__next__()
            except StopIteration:
                return None

        page = await run_in_executor(_get_next_page)
        while page:
            last_evaluated_key = page.get("LastEvaluatedKey", None)
            yield page
            if not last_evaluated_key:
                break
            page = await run_in_executor(_get_next_page)

    async def paginate_items(
        self, paginator_type: Literal["query", "scan", "list_backups", "list_tables"], **kwargs
    ) -> AsyncGenerator[PagedItems[dict], None]:
        async for page in self.paginate(paginator_type, **kwargs):
            items = page.get("Items", [])
            last_evaluated_key = page.get("LastEvaluatedKey", None)
            yield PagedItems(items, last_evaluated_key)

    @dynamodb_retry_backoff()
    async def query(self, **kwargs) -> QueryOutputTableTypeDef:
        return await run_in_executor(self.table.query, **kwargs)

    async def query_items(self, **kwargs) -> PagedItems[dict]:
        page = await self.query(**kwargs)
        return PagedItems(page.get("Items", []), cast(Optional[Dict[str, Any]], page.get("LastEvaluatedKey", None)))

    @dynamodb_retry_backoff()
    async def query_count(self, **kwargs) -> int:
        total = 0
        async for page in self.paginate("query", Select="COUNT", **kwargs):
            page_count = page.get("Count", 0)
            total += page_count

        return total

    @dynamodb_retry_backoff()
    async def transact_write_items(self, transact_items: list, **kwargs):
        return await run_in_executor(self.client.transact_write_items, TransactItems=transact_items, **kwargs)

    def _prepare_transaction(self, actions: List[Dict[str, Any]], table_name: str, **kwargs):
        transact_items = []
        for action in actions:
            for operation, item in action.items():
                assert operation in ("Put", "Update", "Delete", "ConditionCheck"), f"unexpected_operation: {operation}"
                serialised = self.serialise_value(item, **kwargs)
                content: Dict[str, Any] = {"TableName": table_name}
                if operation == "Put" and is_dataclass_instance(item):
                    content["Item"] = serialised
                else:
                    assert isinstance(serialised, dict)
                    content.update(serialised)
                transact_items.append({operation: content})
        return transact_items

    async def transact_write(self, actions: List[Dict[str, Any]]):
        transaction = self._prepare_transaction(actions=actions, table_name=self._table_name)
        return await self.transact_write_items(transaction)

    @dynamodb_retry_backoff()
    async def update_item_with_retry_info(
        self, key: TModelKey, **kwargs
    ) -> Tuple[UpdateItemOutputTableTypeDef, float, str, int]:
        kwargs["Key"] = cast(Dict[str, str], key)

        started = time()
        response = await run_in_executor(self.table.update_item, **kwargs)
        duration = time() - started

        aws_request_id = response["ResponseMetadata"]["RequestId"]
        aws_retries = response["ResponseMetadata"]["RetryAttempts"]

        return response, duration, aws_request_id, aws_retries

    async def update_item(self, key: TModelKey, **kwargs) -> UpdateItemOutputTableTypeDef:
        response, _, _, _ = await self.update_item_with_retry_info(key, **kwargs)
        return response

    def batch_writer(self, flush_amount: int = 25) -> "_AsyncBatchWriter":
        return _AsyncBatchWriter(
            store=self,
            flush_amount=flush_amount,
        )

    async def query_all_items(self, max_items: int = 0, **kwargs) -> List[dict]:
        result = []

        async for items, _ in self.paginate_items("query", **kwargs):
            result.extend(items)
            if 0 < max_items <= len(result):
                return result[:max_items]

        return result

    async def get_all_model_keys(self, model_type: Type[TBaseModel_co], max_keys: int = 0) -> List[TModelKey]:
        assert self._model_type_index_name
        assert self._model_type_index_pk
        results = await self.query_all_items(
            IndexName=self._model_type_index_name,
            KeyConditionExpression=Key(self._model_type_index_pk).eq(model_type.__name__),
            max_items=max_keys,
        )
        return [model_type.model_key_from_item(res) for res in results]

    async def get_all_models(
        self, model_type: Type[TBaseModel_co], max_concurrency=10, max_models: int = 0
    ) -> List[TBaseModel_co]:
        model_keys = await self.get_all_model_keys(model_type=model_type, max_keys=max_models)
        return [
            cast(TBaseModel_co, record)
            for record in await self.batch_get_model(model_keys, max_concurrency=max_concurrency)
            if record
        ]

    async def paginate_models(
        self,
        paginator_type: Literal["query", "scan", "list_backups", "list_tables"],
        model_type: Type[TBaseModel_co],
        **kwargs,
    ) -> AsyncGenerator[PagedItems[TBaseModel_co], None]:
        async for items, last_evaluated_key in self.paginate_items(paginator_type, **kwargs):
            yield PagedItems(
                items=[cast(TBaseModel_co, self.deserialise_model(item, model_type, **kwargs)) for item in items],
                last_evaluated_key=last_evaluated_key,
            )

    async def paginate_models_from_index(
        self,
        paginator_type: Literal["query", "scan", "list_backups", "list_tables"],
        model_type: Type[TBaseModel_co],
        index_name: str,
        **kwargs,
    ) -> AsyncGenerator[PagedItems[TBaseModel_co], None]:
        async for items, last_evaluated_key in self.paginate_items(paginator_type, IndexName=index_name, **kwargs):
            model_keys = [model_type.model_key_from_item(item) for item in items]
            models = await self.batch_get_model(model_keys)
            yield PagedItems(
                [cast(TBaseModel_co, model) for model in models],
                last_evaluated_key=last_evaluated_key,
            )

    async def query_models(self, model_type: Type[TBaseModel_co], **kwargs) -> PagedItems[TBaseModel_co]:
        items, last_evaluated_key = await self.query_items(**kwargs)
        return PagedItems(
            items=[cast(TBaseModel_co, self.deserialise_model(item, model_type, **kwargs)) for item in items],
            last_evaluated_key=last_evaluated_key,
        )

    async def query_all_models(self, model_type: Type[TBaseModel_co], **kwargs) -> List[TBaseModel_co]:
        result = []

        async for models, _ in self.paginate_models("query", model_type, **kwargs):
            result.extend(models)
        return result

    def _get_batch_results(
        self, response: BatchGetItemOutputServiceResourceTypeDef
    ) -> Tuple[List[Dict[str, Any]], List[TModelKey]]:
        assert response["ResponseMetadata"]["HTTPStatusCode"] == 200
        results: List[Dict[str, Any]] = response.get("Responses", {}).get(self._table_name) or []
        unprocessed_keys: Dict[str, Any] = response.get("UnprocessedKeys", {})
        unprocessed = cast(List[TModelKey], unprocessed_keys.get(self._table_name, {}).get("Keys") or [])

        return results, unprocessed

    @dynamodb_retry_backoff()
    async def _get_batch(self, keys: List[TModelKey], **kwargs) -> Tuple[List[Dict[str, Any]], List[TModelKey]]:
        if not keys:
            return [], []

        request = kwargs or {}
        request["Keys"] = keys

        response = await run_in_executor(self.service.batch_get_item, RequestItems={self._table_name: request})
        return self._get_batch_results(response)

    async def batch_get_item(self, keys: List[TModelKey], max_concurrency: int = 10, **kwargs) -> List[Dict[str, Any]]:
        def _chunk(it, size):
            it = iter(it)
            return iter(lambda: tuple(itertools.islice(it, size)), ())

        result = []
        remaining = keys
        while remaining:
            batches = _chunk(remaining, 100)  # into batches of max size 100 (max allowed)
            to_retry = []
            async with asyncio.Semaphore(max_concurrency):
                task_results = await asyncio.gather(*[self._get_batch(list(batch), **kwargs) for batch in batches])

            for found, unprocessed in task_results:
                if found:
                    result.extend(found)
                if unprocessed:
                    to_retry.extend(unprocessed)

            remaining = to_retry

        return result

    async def batch_get_item_ordered(
        self, keys: Sequence[TModelKey], max_concurrency: int = 10, **kwargs
    ) -> List[Optional[dict]]:
        keys = list(keys)
        key_indexes: Dict[Tuple[str, ...], int] = {self._model_key_tuple(key): ix for ix, key in enumerate(keys)}

        records = await self.batch_get_item(keys, max_concurrency=max_concurrency, **kwargs)

        results: List[Optional[dict]] = [None for _ in range(len(keys))]

        unexpected = []

        for record in records:
            key_tuple = self._model_key_tuple(record)
            ix = key_indexes.get(key_tuple)
            if ix is None:
                unexpected.append(key_tuple)
                continue
            results[ix] = record

        if unexpected:
            raise ValueError(f"unexpected items in batch result: {list(unexpected)}")

        return results

    async def batch_get_model(
        self, keys: Sequence[TModelKey], max_concurrency: int = 10, **kwargs
    ) -> List[Optional[TBaseModel]]:
        ordered = await self.batch_get_item_ordered(keys, max_concurrency=max_concurrency)

        models: List[Optional[TBaseModel]] = [
            (
                cast(TBaseModel, self.deserialise_model(record, self._model_types[record["model_type"]], **kwargs))
                if record
                else None
            )
            for record in ordered
        ]

        return models


class _AsyncBatchWriter:
    """
    converted from boto3.dynamodb.BatchWriter to be async
    """

    def __init__(self, store: BaseModelStore, flush_amount=25):
        self._store = store
        self._table_name = store.table.table_name
        self._client = store.client
        self._items_buffer: List[WriteRequestOutputTypeDef] = []
        self._flush_amount = flush_amount
        self._overwrite_by_keys = store.table_key_fields

    async def put_item(self, item, **kwargs):
        await self._add_request_and_process({"PutRequest": {"Item": self._store.serialise_value(item, **kwargs)}})

    async def delete_item(self, key, **kwargs):
        await self._add_request_and_process({"DeleteRequest": {"Key": self._store.serialise_value(key, **kwargs)}})

    async def _add_request_and_process(self, request):
        if self._overwrite_by_keys:
            self._remove_dup_keys_request_if_any(request)
        self._items_buffer.append(request)
        await self._flush_if_needed()

    def _remove_dup_keys_request_if_any(self, request):
        pkey_values_new = self._extract_pkey_values(request)
        for item in self._items_buffer:
            if self._extract_pkey_values(item) == pkey_values_new:
                self._items_buffer.remove(item)

    def _extract_pkey_values(self, request):
        if request.get("PutRequest"):
            return [request["PutRequest"]["Item"][key] for key in self._overwrite_by_keys]
        if request.get("DeleteRequest"):
            return [request["DeleteRequest"]["Key"][key] for key in self._overwrite_by_keys]
        return None

    async def _flush_if_needed(self):
        if len(self._items_buffer) >= self._flush_amount:
            await self._flush()

    @dynamodb_retry_backoff()
    async def _flush(self):
        if not self._items_buffer:
            return

        items_to_send = self._items_buffer[: self._flush_amount]
        if not items_to_send:
            return

        self._items_buffer = self._items_buffer[self._flush_amount :]
        response = await run_in_executor(self._client.batch_write_item, RequestItems={self._table_name: items_to_send})
        unprocessed_items = response["UnprocessedItems"]

        if unprocessed_items and unprocessed_items[self._table_name]:
            # Any unprocessed_items are immediately added to the
            # next batch we send.
            self._items_buffer.extend(unprocessed_items[self._table_name])
        else:
            self._items_buffer = []

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_value, tb):
        # When we exit, we need to keep flushing whatever is left
        # until there's nothing left in our items buffer.
        if not exc_value:
            while len(self._items_buffer) > 0:
                await self._flush()
        return True



import pickle
import random
import sys
from dataclasses import dataclass, field
from datetime import date, datetime
from enum import Enum
from typing import Any, Dict, Generator, List, Mapping, Optional, Type, TypedDict, Union, cast
from uuid import uuid4

import petname  # type: ignore[import]
import pytest as pytest
from boto3.dynamodb.conditions import Key
from mypy_boto3_dynamodb.service_resource import Table

from nhs_aws_helpers import dynamodb
from nhs_aws_helpers.dynamodb_model_store.base_model import (
    BaseModel,
    serialised_property,
)
from nhs_aws_helpers.dynamodb_model_store.base_model_store import (
    BaseModelStore,
    PagedItems,
)


class _MyModelKey(TypedDict):
    my_pk: str
    my_sk: str


@pytest.mark.skipif(sys.version_info < (3, 10), reason="requires python3.10 or higher")
def test_create_py_310_model():
    @dataclass(kw_only=True)  # type: ignore[call-overload]
    class MyBaseModelPy310(BaseModel[_MyModelKey]):
        _model_key_type = _MyModelKey
        base_required_thing: str
        optional_thing: Optional[str] = None

        last_modified: Optional[datetime] = field(default_factory=datetime.utcnow)

        def __post_init__(self):
            self.optional_thing = self.optional_thing or uuid4().hex

    @dataclass
    class SomeDerivedModelPy310(MyBaseModelPy310):
        _index_model_type = False
        required_thing: str  # type: ignore[misc]
        another_optional_thing: Optional[bool] = field(default=None)

        def _create_key(self) -> _MyModelKey:
            return self.key_from(self.required_thing)

        @classmethod
        def key_from(cls, required_thing: str) -> _MyModelKey:
            return _MyModelKey(my_pk="TEST", my_sk=required_thing)

        def __post_init__(self):
            # call this is you have a base __post_init__
            super().__post_init__()
            self.another_optional_thing = True

    model = SomeDerivedModelPy310(base_required_thing="TEST", required_thing="TEST2")
    assert model.last_modified
    assert model.another_optional_thing
    assert model.optional_thing
    assert model.base_required_thing == "TEST"
    assert model.required_thing == "TEST2"
    assert MyBaseModelPy310.model_key_fields() == ["my_pk", "my_sk"]
    assert SomeDerivedModelPy310.model_key_fields() == ["my_pk", "my_sk"]


def test_create_model_store_without_base_model_type():
    class BadModelStore(BaseModelStore[BaseModel, _MyModelKey]):
        pass

    with pytest.raises(ValueError, match=f"{BadModelStore.__name__} needs a _base_model_type definition"):
        BadModelStore("no")


class MyBaseModel(BaseModel[_MyModelKey]):
    """
    pre py 3.10 dataclass inheritance doesn't work very well ... unless al
    but you may still need a base model
    """

    _model_key_type = _MyModelKey

    def get_key(self) -> _MyModelKey:
        raise NotImplementedError("should be implemented in the dervied class")

    @serialised_property
    def my_pk(self) -> str:
        return self.get_key()["my_pk"]

    @serialised_property
    def my_sk(self) -> str:
        return self.get_key()["my_sk"]


@dataclass
class NestedItem:
    event: str
    timestamp: datetime = field(default_factory=datetime.utcnow)


class SomeEnum(str, Enum):
    FIELD_ONE = "one"
    FIELD_TWO = "two"


@dataclass
class NestedModelWithEnum:
    some_enum: SomeEnum
    some_str: str


@dataclass
class MyDerivedModel(MyBaseModel):
    id: str
    sk_field: Optional[str] = None
    optional_thing: Optional[str] = None

    today: date = field(default_factory=lambda: datetime.utcnow().date())
    last_modified: datetime = field(default_factory=datetime.utcnow)
    some_dict: Optional[dict] = field(default_factory=lambda: {"key": 123})
    chaos: Optional[int] = field(default_factory=lambda: random.randint(0, 10))
    set_type: Optional[set] = field(default_factory=set)
    frozenset_type: Optional[frozenset] = field(default_factory=frozenset)
    none_string: Optional[str] = None
    none_list: Optional[List[str]] = None
    nested_items: List[NestedItem] = field(default_factory=list)
    nested_item: NestedItem = field(default_factory=lambda: NestedItem(event="created"))
    union_date: Union[datetime, None] = field(default_factory=datetime.utcnow)
    bytes_type: Union[bytes, None] = None
    bytearray_type: Union[bytearray, None] = None
    some_enum: SomeEnum = SomeEnum.FIELD_ONE
    nested_enum: NestedModelWithEnum = field(
        default_factory=lambda: NestedModelWithEnum(some_str="string", some_enum=SomeEnum.FIELD_TWO)
    )

    def get_key(self) -> _MyModelKey:
        return _MyModelKey(my_pk=f"AA#{self.id}", my_sk=self.sk_field or "#")

    def __post_init__(self):
        self.sk_field = self.sk_field or "#"
        if (self.chaos or 0) > 4:
            if not self.set_type:
                self.set_type = {self.chaos}
            if not self.frozenset_type:
                self.frozenset_type = frozenset(self.set_type)


@dataclass
class AnotherModel(MyBaseModel):
    id: str
    sk_field: Optional[str] = None
    last_modified: datetime = field(default_factory=datetime.utcnow)

    def get_key(self) -> _MyModelKey:
        return _MyModelKey(my_pk=f"BB#{self.id}", my_sk=self.sk_field or "#")

    def __post_init__(self):
        self.sk_field = self.sk_field or "#"


@dataclass
class UnregisteredModel(MyBaseModel):
    id: str
    sk_field: Optional[str] = None
    last_modified: datetime = field(default_factory=datetime.utcnow)

    def get_key(self) -> _MyModelKey:
        return _MyModelKey(my_pk=f"AA#{self.id}", my_sk=self.sk_field or "#")

    def __post_init__(self):
        self.sk_field = self.sk_field or "#"


class NotSerializable:
    def __init__(self, some_uuid: str):
        self._some_uuid = some_uuid

    @property
    def some_uuid(self) -> str:
        return self._some_uuid


@dataclass
class ASpecialDerivedModel(MyBaseModel):
    id: str
    special_field: NotSerializable
    events: List[NestedItem] = field(default_factory=list)
    last_modified: datetime = field(default_factory=datetime.utcnow)

    def get_key(self) -> _MyModelKey:
        return _MyModelKey(my_pk=f"BB#{self.id}", my_sk="#")

    def __post_init__(self):
        if not self.events:
            self.events.append(NestedItem(event="created"))


def test_create_model():
    model = MyDerivedModel(id="TEST2")
    assert model.last_modified
    assert model.sk_field == "#"
    assert model.id == "TEST2"
    assert MyBaseModel.model_key_fields() == ["my_pk", "my_sk"]
    assert MyDerivedModel.model_key_fields() == ["my_pk", "my_sk"]
    assert MyDerivedModel.model_key_fields() == ["my_pk", "my_sk"]
    assert MyBaseModel.model_key_fields() == ["my_pk", "my_sk"]


class MyModelStore(BaseModelStore[MyBaseModel, _MyModelKey]):
    _base_model_type = MyBaseModel
    _model_types = {cls.__name__: cls for cls in [MyDerivedModel, AnotherModel, ASpecialDerivedModel]}  # noqa: RUF012

    def __init__(self, table_name: str):
        super().__init__(table_name, model_type_index_name="gsi_model_type", model_type_index_pk="model_type")
        self.register_model_type(MyDerivedModel)

    @classmethod
    def _before_deserialise_model(
        cls, model_dict: Dict[str, Any], model_type: Type[MyBaseModel], **kwargs
    ) -> Dict[str, Any]:
        if not model_dict:
            return model_dict

        if model_type != ASpecialDerivedModel or "special_field" not in model_dict:
            return model_dict

        model_dict["special_field"] = pickle.loads(model_dict["special_field"].value)
        return model_dict

    @classmethod
    def _serialise_field(
        cls, model: Any, field_name: str, field_type: type, metadata: Mapping[str, Any], value: Any, **kwargs
    ) -> Any:
        if value is None:
            return None

        if field_type != NotSerializable:
            return super()._serialise_field(model, field_name, field_type, metadata, value, **kwargs)

        return pickle.dumps(value)


@pytest.fixture(scope="session", name="session_temp_ddb_table")
def create_session_temp_ddb_table() -> Generator[Table, None, None]:
    ddb = dynamodb()

    args = {
        "AttributeDefinitions": [
            {"AttributeName": "my_pk", "AttributeType": "S"},
            {"AttributeName": "my_sk", "AttributeType": "S"},
            {"AttributeName": "model_type", "AttributeType": "S"},
        ],
        "KeySchema": [{"AttributeName": "my_pk", "KeyType": "HASH"}, {"AttributeName": "my_sk", "KeyType": "RANGE"}],
        "BillingMode": "PAY_PER_REQUEST",
        "GlobalSecondaryIndexes": [
            {
                "IndexName": "gsi_model_type",
                "KeySchema": [
                    {"AttributeName": "model_type", "KeyType": "HASH"},
                ],
                "Projection": {
                    "ProjectionType": "KEYS_ONLY",
                },
            },
        ],
    }

    table_name = f"pytest-{petname.Generate(words=4, separator='_')}"

    table = ddb.create_table(TableName=table_name, **args)  # type: ignore[arg-type]

    yield table

    table.delete()


@pytest.fixture(name="temp_table")
def create_temp_ddb_table(session_temp_ddb_table: Table) -> Table:
    table = session_temp_ddb_table
    result = table.scan(ProjectionExpression="my_pk, my_sk", ConsistentRead=True)
    with table.batch_writer(["my_pk", "my_sk"]) as writer:
        while True:
            items = result.get("Items", [])
            if not items:
                break
            for item in items:
                writer.delete_item({"my_pk": item["my_pk"], "my_sk": item["my_sk"]})
            if len(writer._items_buffer) > 0:  # type: ignore[attr-defined]
                writer._flush()  # type: ignore[attr-defined]
            if not result.get("LastEvaluatedKey"):
                break

    return table


@pytest.fixture(name="store")
def create_temp_store(temp_table: Table):
    return MyModelStore(temp_table.name)


async def test_store_get_item(store: MyModelStore):
    pk = uuid4().hex
    item = {"my_pk": pk, "my_sk": "#", "field": uuid4().hex}
    store.table.put_item(Item=item)

    result = await store.get_item(_MyModelKey(my_pk=pk, my_sk="#"))
    assert result
    assert result["field"] == item["field"]


async def test_store_put_get_item(store: MyModelStore):
    pk = uuid4().hex
    item = {"my_pk": pk, "my_sk": "#", "field": uuid4().hex}

    await store.put_item(item)

    result = await store.get_item(_MyModelKey(my_pk=pk, my_sk="#"))
    assert result
    assert result["field"] == item["field"]


async def test_store_put_get_model(store: MyModelStore):
    expected = MyDerivedModel(id=uuid4().hex)

    await store.put_model(expected)

    actual = await store.get_model(expected.model_key(), MyDerivedModel)

    assert actual
    assert isinstance(actual, MyDerivedModel)


async def test_store_put_update_get_model(store: MyModelStore):
    expected = MyDerivedModel(id=uuid4().hex)

    await store.put_model(expected)

    actual = await store.get_model(expected.model_key(), MyDerivedModel)

    assert actual
    assert isinstance(actual, MyDerivedModel)

    await store.update_item(
        expected.model_key(), UpdateExpression="SET optional_thing = :val", ExpressionAttributeValues={":val": "BOB"}
    )

    actual = await store.get_model(expected.model_key(), MyDerivedModel)
    assert actual
    assert actual.optional_thing == "BOB"


async def test_store_put_exists_delete_exists(store: MyModelStore):
    expected = MyDerivedModel(id=uuid4().hex)

    await store.put_model(expected)

    model = await store.get_model(expected.model_key(), MyDerivedModel)

    assert model

    exists = await store.item_exists(expected.model_key())
    assert exists

    await store.delete_item(expected.model_key())

    actual = await store.get_item(expected.model_key())

    assert actual is None

    exists = await store.item_exists(expected.model_key())
    assert not exists


async def test_store_put_model_if_not_exists(store: MyModelStore):
    original = MyDerivedModel(id=uuid4().hex, optional_thing=uuid4().hex)

    was_put = await store.put_model_if_not_exists(original)
    assert was_put

    new = MyDerivedModel(id=original.id, optional_thing=uuid4().hex)

    was_put = await store.put_model_if_not_exists(new)
    assert not was_put

    got = await store.get_model(new.model_key(), MyDerivedModel)
    assert got
    assert got.optional_thing == original.optional_thing


async def test_store_put_item_if_not_exists(store: MyModelStore):
    original = {"my_pk": uuid4().hex, "my_sk": "A", "field": uuid4().hex}

    was_put = await store.put_item_if_not_exists(original)
    assert was_put

    new = {"my_pk": original["my_pk"], "my_sk": "A", "field": uuid4().hex}

    was_put = await store.put_item_if_not_exists(new)
    assert not was_put

    got = await store.get_item(_MyModelKey(my_pk=new["my_pk"], my_sk=new["my_sk"]))
    assert got
    assert got["field"] == original["field"]


async def test_inject_custom_serialization(store: MyModelStore):
    put_model = ASpecialDerivedModel(id=uuid4().hex, special_field=NotSerializable(uuid4().hex))

    await store.put_model(put_model)

    got_model = await store.get_model(put_model.model_key(), ASpecialDerivedModel)

    assert got_model
    assert got_model.special_field is not None
    assert got_model.special_field.some_uuid == put_model.special_field.some_uuid
    assert got_model.events
    assert len(got_model.events) == 1
    assert isinstance(got_model.events[0], NestedItem)
    assert got_model.events[0].timestamp == put_model.events[0].timestamp


async def test_query_and_unpack_paged_items(store: MyModelStore):
    partition_key = "PK#1"

    async with store.batch_writer() as writer:
        for i in range(30):
            await writer.put_item({"my_pk": partition_key, "my_sk": f"SK#{i}"})

    items, last_evaluated_key = await store.query_items(KeyConditionExpression=Key("my_pk").eq(partition_key))

    assert len(items) == 30
    assert last_evaluated_key is None


async def test_query_and_unpack_paged_items_with_limit(store: MyModelStore):
    partition_key = "PK#1"

    async with store.batch_writer() as writer:
        for i in range(20):
            await writer.put_item({"my_pk": partition_key, "my_sk": f"SK#{i}"})

    items, last_evaluated_key = await store.query_items(KeyConditionExpression=Key("my_pk").eq(partition_key), Limit=10)

    assert len(items) == 10
    assert last_evaluated_key is not None


async def test_query_paginate_items(store: MyModelStore):
    partition_key = "PK#1"

    async with store.batch_writer() as writer:
        for i in range(20):
            await writer.put_item({"my_pk": partition_key, "my_sk": f"SK#{i}"})

    pages = [
        page
        async for page in store.paginate_items("query", KeyConditionExpression=Key("my_pk").eq(partition_key), Limit=10)
    ]

    assert len(pages) == 3
    assert len(pages[0].items) == 10
    assert len(pages[1].items) == 10
    assert len(pages[2].items) == 0
    assert pages[0].last_evaluated_key is not None
    assert pages[1].last_evaluated_key is not None
    assert pages[2].last_evaluated_key is None


async def test_try_deserialise_non_model(store: MyModelStore):
    with pytest.raises(TypeError):
        store.deserialise_model({"aa": 123}, NotSerializable)  # type: ignore[type-var]


async def test_serialize_deserialize_model(store: MyModelStore):
    model = MyDerivedModel(id=uuid4().hex, nested_items=[NestedItem(event="bob"), NestedItem(event="fin")])
    serialized = store.serialise_model(model)

    assert serialized
    assert serialized["my_pk"] == model.my_pk
    assert serialized["my_sk"] == model.my_sk
    assert serialized["model_type"] == MyDerivedModel.__name__
    assert serialized["id"] == model.id
    assert model.some_dict
    assert serialized["some_dict"] == model.some_dict
    assert serialized["last_modified"] == model.last_modified.isoformat()
    assert serialized["today"] == datetime.utcnow().date().isoformat()
    assert serialized["chaos"] == model.chaos
    if model.set_type:
        assert serialized["set_type"] == model.set_type
    else:
        assert "set_type" not in serialized

    if model.frozenset_type:
        assert serialized["frozenset_type"] == model.frozenset_type
    else:
        assert "frozenset_type" not in serialized

    assert len(serialized["nested_items"]) == 2
    assert isinstance(serialized["nested_items"][0], dict)
    assert isinstance(serialized["nested_item"], dict)
    assert serialized["nested_item"]["event"] == "created"

    assert isinstance(serialized["some_enum"], str)
    assert serialized["some_enum"] == "one"
    assert isinstance(serialized["nested_enum"], dict)
    assert serialized["nested_enum"]["some_enum"] == "two"

    assert model.none_string is None
    assert "none_thing" not in serialized
    assert model.none_list is None
    assert "none_list" not in serialized

    deserialized = store.deserialise_model(serialized, MyDerivedModel)
    assert deserialized
    assert deserialized.model_key() == model.model_key()
    assert deserialized.id == model.id
    assert deserialized.chaos == model.chaos
    assert len(deserialized.nested_items) == 2
    assert isinstance(deserialized.nested_items[0], NestedItem)
    assert isinstance(deserialized.nested_item, NestedItem)
    assert deserialized.nested_item.event == "created"
    assert deserialized.last_modified == model.last_modified
    assert deserialized.today == model.today
    assert isinstance(deserialized.some_enum, SomeEnum)
    assert deserialized.some_enum == SomeEnum.FIELD_ONE
    assert isinstance(deserialized.nested_enum, NestedModelWithEnum)
    assert isinstance(deserialized.nested_enum.some_enum, SomeEnum)
    assert deserialized.nested_enum.some_enum == SomeEnum.FIELD_TWO


async def test_nested_enum_using_deserialise_model_function():
    enum_test_dict = {
        "some_str": "test string",
        "some_enum": "two",
    }

    deserialised_model = BaseModelStore.deserialise_model(enum_test_dict, NestedModelWithEnum)  # type: ignore[type-var]

    assert isinstance(deserialised_model, NestedModelWithEnum)
    assert isinstance(deserialised_model.some_enum, SomeEnum)
    assert deserialised_model.some_enum == SomeEnum.FIELD_TWO
    assert deserialised_model.some_str == "test string"


async def test_transact_get_put_model(store: MyModelStore):
    model = MyDerivedModel(id=uuid4().hex)
    await store.transact_write(actions=[{"Put": model}])

    got = await store.transact_get_model(model.model_key(), MyDerivedModel)

    assert got
    assert got.id == model.id
    assert got.union_date is not None
    assert isinstance(got.union_date, datetime)


async def test_transact_get_models(store: MyModelStore):
    models = [
        MyDerivedModel(id=uuid4().hex),
        ASpecialDerivedModel(id=uuid4().hex, special_field=NotSerializable(uuid4().hex)),
    ]

    await store.transact_write(actions=[{"Put": model} for model in models])

    keys = [
        models[-1].model_key(),
        _MyModelKey(my_pk=uuid4().hex, my_sk=uuid4().hex),
        _MyModelKey(my_pk=uuid4().hex, my_sk=uuid4().hex),
        models[0].model_key(),
    ]

    results = await store.transact_get_models(keys)

    assert len(results) == 4
    assert results[1] is None
    assert results[2] is None
    assert isinstance(results[0], ASpecialDerivedModel)
    assert isinstance(results[-1], MyDerivedModel)


async def test_query_all_items(store: MyModelStore):
    partition_key = "PK:1"

    async with store.batch_writer() as writer:
        for i in range(20):
            await writer.put_item({"my_pk": partition_key, "my_sk": f"SK:{i}"})

    items = await store.query_all_items(KeyConditionExpression=Key("my_pk").eq(partition_key))

    assert len(items) == 20

    items = await store.query_all_items(KeyConditionExpression=Key("my_pk").eq(partition_key), Limit=20)
    assert len(items) == 20


async def test_query_all_models(store: MyModelStore):
    async with store.batch_writer() as writer:
        for i in range(20):
            await writer.put_item(MyDerivedModel(id="1", sk_field=f"SK:{i}"))

    models = await store.query_all_models(MyDerivedModel, KeyConditionExpression=Key("my_pk").eq("AA#1"))

    assert len(models) == 20

    models = await store.query_all_models(MyDerivedModel, KeyConditionExpression=Key("my_pk").eq("AA#1"), Limit=10)
    assert len(models) == 20

    page = await store.query_models(MyDerivedModel, KeyConditionExpression=Key("my_pk").eq("AA#1"))
    assert len(page) == 20


async def test_batch_get_model(store: MyModelStore):
    models: List[MyBaseModel] = [MyDerivedModel(id=uuid4().hex) for _ in range(10)]
    models.extend([AnotherModel(id=uuid4().hex) for _ in range(10)])
    async with store.batch_writer() as writer:
        for model in models:
            await writer.put_item(model)

    got = await store.batch_get_model([model.model_key() for model in models])

    assert len(got) == 20
    assert all(isinstance(model, (MyDerivedModel, AnotherModel)) for model in got)


async def test_batch_get_item(store: MyModelStore):
    models: List[MyBaseModel] = [MyDerivedModel(id=uuid4().hex) for _ in range(10)]
    models.extend([AnotherModel(id=uuid4().hex) for _ in range(10)])
    async with store.batch_writer() as writer:
        for model in models:
            await writer.put_item(model)

    got = await store.batch_get_item([model.model_key() for model in models], ProjectionExpression="last_modified")

    assert len(got) == 20
    assert all(set(item.keys()) == {"last_modified"} for item in got)


async def test_batch_get_item_ordered(store: MyModelStore):
    models: List[MyBaseModel] = [MyDerivedModel(id=uuid4().hex) for _ in range(10)]
    models.extend([AnotherModel(id=uuid4().hex) for _ in range(10)])
    async with store.batch_writer() as writer:
        for model in models:
            await writer.put_item(model)

    got = cast(
        List[dict],
        await store.batch_get_item_ordered(
            [model.model_key() for model in models], ProjectionExpression="my_pk,my_sk,last_modified"
        ),
    )

    assert len(got) == 20
    assert all(set(item.keys()) == {"my_pk", "my_sk", "last_modified"} for item in got)


async def test_unregistered_model(store: MyModelStore):
    models: List[MyBaseModel] = [UnregisteredModel(id=uuid4().hex) for _ in range(10)]
    async with store.batch_writer() as writer:
        for model in models:
            await writer.put_item(model)

    with pytest.raises(KeyError):
        await store.batch_get_model([model.model_key() for model in models])


async def test_query_count(store: MyModelStore):
    partition_key = "PK:1"

    async with store.batch_writer() as writer:
        for i in range(100):
            await writer.put_item({"my_pk": partition_key, "my_sk": f"SK:{i}"})

    count = await store.query_count(KeyConditionExpression=Key("my_pk").eq(partition_key))

    assert count == 100

    count = await store.query_count(KeyConditionExpression=Key("my_pk").eq(partition_key), Limit=7)

    assert count == 100


async def test_get_all_models(store: MyModelStore):
    async with store.batch_writer() as writer:
        for _i in range(100):
            await writer.put_item(AnotherModel(uuid4().hex))

    models = await store.get_all_models(AnotherModel)

    assert len(models) == 100
    assert all(isinstance(model, AnotherModel) for model in models)


async def test_get_all_models_with_a_limit(store: MyModelStore):
    async with store.batch_writer() as writer:
        for _i in range(100):
            await writer.put_item(AnotherModel(uuid4().hex))

    models = await store.get_all_models(AnotherModel, max_models=10)

    assert len(models) == 10
    assert all(isinstance(model, AnotherModel) for model in models)


async def test_paginate_models(store: MyModelStore):
    partition_key = "BB#1"

    async with store.batch_writer() as writer:
        for i in range(100):
            await writer.put_item(AnotherModel(id="1", sk_field=f"SK:{i}"))

    pages = [
        page
        async for page in store.paginate_models(
            "query", AnotherModel, KeyConditionExpression=Key("my_pk").eq(partition_key), Limit=40
        )
    ]

    assert len(pages) == 3


async def test_paginate_models_from_index(store: MyModelStore):
    async with store.batch_writer() as writer:
        for i in range(100):
            await writer.put_item(AnotherModel(id="1", sk_field=f"SK:{i}"))

    pages = [
        page
        async for page in store.paginate_models_from_index(
            paginator_type="query",
            model_type=AnotherModel,
            index_name="gsi_model_type",
            KeyConditionExpression=Key("model_type").eq(AnotherModel.__name__),
            Limit=40,
        )
    ]

    assert len(pages) == 3
    for page in pages:
        assert all(isinstance(model, AnotherModel) for model in page.items)


async def test_paged_items():
    has_size = PagedItems(items=[1, 2, 3], last_evaluated_key={"my_pk": 123})
    assert has_size
    assert len(has_size) == 3

    empty_items: PagedItems = PagedItems(items=[], last_evaluated_key=None)
    assert not empty_items
    assert len(empty_items) == 0

    none_items: PagedItems = PagedItems(items=None, last_evaluated_key=None)  # type: ignore[arg-type]
    assert not none_items
    assert len(none_items) == 0

    has_key: PagedItems = PagedItems(items=[], last_evaluated_key={"my_pk": 456})
    assert not has_key
    assert len(has_key) == 0

    summed: PagedItems = PagedItems(items=[], last_evaluated_key=None)
    for page in [has_size, empty_items, none_items, has_key]:
        summed += page

    assert len(summed.items) == 3
    assert summed.last_evaluated_key == has_key.last_evaluated_key

    new_items = has_size + has_size
    assert len(new_items.items) == 6


async def test_bytes_serialisation(store: MyModelStore):
    bytes_in = b"test"
    model = MyDerivedModel(id=uuid4().hex, bytes_type=bytes_in)
    await store.put_model(model)
    stored = await store.get_model(model.model_key(), MyDerivedModel)
    assert stored
    assert isinstance(stored.bytes_type, bytes)
    assert stored.bytes_type == bytes_in


async def test_bytearray_serialisation(store: MyModelStore):
    bytes_in = bytearray(b"test")
    model = MyDerivedModel(id=uuid4().hex, bytearray_type=bytes_in)
    await store.put_model(model)
    stored = await store.get_model(model.model_key(), MyDerivedModel)
    assert stored
    assert isinstance(stored.bytearray_type, bytearray)
    assert stored.bytearray_type == bytes_in
